# Copyright (C) 2011 Oracle. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, version 2.  This program is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the GNU
# General Public License along with this program; if not, write to the Free
# Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 021110-1307, USA.

import logging
import os
import re
import subprocess
import sys


BASEDIR = '/etc/template.d'
TARGETS = ['configure', 'unconfigure', 'reconfigure', 'cleanup', 'suspend',
           'resume', 'migrate', 'shutdown']


def get_target_path(target):
    return os.path.join(BASEDIR, '%s.d' % target)


def get_script_dir():
    return os.path.join(BASEDIR, 'scripts')


def get_script_path(name):
    return os.path.join(BASEDIR, 'scripts', name)


def check_name(name):
    if not os.path.isfile('/etc/template.d/scripts/%s' % name):
        error('script does not exist: %s', name)


def get_name(lnk):
    lnk_name = os.path.basename(lnk)
    match = re.match(r'^\d\d(?P<name>[a-zA-Z0-9_-]+)$', lnk_name)
    if not match:
        error('cannot get script name from link: %s', lnk)
    return match.group('name')


def check_priority(priority):
    if not re.match(r'^\d\d$', priority):
        error('unrecognized priority: %s', priority)


def get_priority(lnk):
    lnk_name = os.path.basename(lnk)
    match = re.match(r'^(?P<priority>\d\d)[a-zA-Z0-9_-]+$', lnk_name)
    if not match:
        error('cannot get priority from link: %s', lnk)
    return match.group('priority')


def msg(message, *args):
    message = str(message)
    if args:
        message = message % args
    sys.stdout.write(message)
    sys.stdout.flush()


def error(message, *args):
    message = str(message)
    if args:
        message = message % args
    print >> sys.stderr, message
    sys.exit(1)


def get_entry_list(param, entry):
    entry_list = []
    entry_re = re.compile(r'^%s.(?P<index>\d+)$' % entry)
    for (key, value) in sorted(param.items()):
        entry_match = entry_re.match(key)
        if entry_match:
            index = entry_match.group('index')
            value = value.strip()
            if value:
                entry_list.append((value, index))
    return entry_list


def run_cmd(cmd, inputs=None):
    logging.debug('run cmd: %s', cmd)
    proc = subprocess.Popen(cmd, stdin=subprocess.PIPE, stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE, close_fds=True)
    (stdoutdata, stderrdata) = proc.communicate(inputs)
    if proc.returncode != 0:
        raise RuntimeError('command failed: cmd (%s) rc (%s) stdout (%s) '
                           'stderr (%s)' % (cmd, proc.returncode, stdoutdata,
                           stderrdata))
    return stdoutdata


def shell_cmd(cmd):
    logging.debug('shell cmd: %s', cmd)
    proc = subprocess.Popen(cmd, stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE, shell=True, close_fds=True)
    (stdoutdata, stderrdata) = proc.communicate()
    if proc.returncode != 0:
        logging.debug('shell command failed: cmd (%s) rc (%s) stdout (%s) stderr (%s)',
                      cmd, proc.returncode, stdoutdata, stderrdata)
    return (proc.returncode, stdoutdata, stderrdata)


def pipeline(cmds, outputfile=None, success_return_codes=(0,)):
    logging.debug('pipeline: %s', cmds)
    if not cmds:
        raise RuntimeError('Empty command list.')
    output = None
    if outputfile:
        output = open(outputfile, 'w')
    try:
        proc = None
        for cmd in cmds:
            if proc:
                stdin = proc.stdout
            else:
                stdin = None
            if output and cmd == cmds[-1]:
                stdout = output
            else:
                stdout = subprocess.PIPE
            if not isinstance(cmd, list):
                raise RuntimeError('Only list is accepted as command args.')
            cmd = [str(x) for x in cmd]
            new_proc = subprocess.Popen(cmd, stdin=stdin, stdout=stdout,
                                        stderr=subprocess.PIPE, close_fds=True)
            if proc:
                proc.stdout.close()
            proc = new_proc
        (stdoutdata, stderrdata) = proc.communicate()
        if proc.returncode not in success_return_codes:
            raise RuntimeError('Command failed: cmds (%s) output file (%s) '
                      'success_return_codes %s) rc (%s) stdout (%s) stderr '
                      '(%s)' % (cmds, outputfile, success_return_codes,
                      proc.returncode, stdoutdata, stderrdata))
        return stdoutdata
    finally:
        if output:
            output.close()


def set_password(username, password):
    if not password:
        raise Exception('empty password!')
    run_cmd(['passwd', '--stdin', username], password)


def check_password(password):
    output = pipeline([['echo', password], ['cracklib-check']])
    match = re.match(r'%s: (?P<result>.*)' % re.escape(password), output)
    if not match:
        raise ValueError('check failed')
    result = match.group('result')
    if not result == 'OK':
        raise ValueError(result)
