# Copyright (C) 2011 Oracle. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, version 2.  This program is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the GNU
# General Public License along with this program; if not, write to the Free
# Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 021110-1307, USA.

import os
import re
import socket
import struct

from os.path import join, exists, isfile, isdir, basename

from templateconfig.conf import ConfShellVar
from templateconfig.common import shell_cmd


SYSCONFDEVICEDIR = "/etc/sysconfig/network-scripts/"
HOSTNAMEFILE = "/etc/hostname"
HOSTSFILE = "/etc/hosts"
SYSNETDIR = "/sys/class/net/"
BONDING_MASTERS = "/sys/class/net/bonding_masters"


class NetworkError(Exception):
    pass


class ConfNetwork(ConfShellVar):
    def __init__(self, filename):
        ConfShellVar.__init__(self, filename)
        if not os.access(self.filename, os.R_OK):
            self.oldmode = 0644
        else:
            status = os.stat(self.filename)
            self.oldmode = status[0]

    def write(self):
        self.chmod(self.oldmode)
        ConfShellVar.write(self)

    def get(self, key, default=None):
        if key in self.vars:
            return self[key]
        return default

    def update(self, newdict=None, **kwargs):
        if newdict is None:
            pass
        else:
            for (k, v) in newdict.items():
                self[k] = v
        if len(kwargs):
            self.vars.update(kwargs)

    def remove(self):
        if exists(self.filename):
            os.unlink(self.filename)

    def clear(self):
        for key in self.keys():
            del self[key]
        self.lines = []


class ConfDevice(ConfNetwork):
    def __init__(self, name):
        ConfNetwork.__init__(self, SYSCONFDEVICEDIR + "ifcfg-" + name)


class ConfRoute(ConfNetwork):
    def __init__(self, name):
        ConfNetwork.__init__(self, SYSCONFDEVICEDIR + "route-" + name)


class ConfMeta(ConfNetwork):
    def __init__(self, name):
        ConfNetwork.__init__(self, SYSCONFDEVICEDIR + "meta-" + name)


def getConfiguredDevices():
    # ignored files: ifcfg- | *~ | *.bak | *.orig | *.rpmnew | *.rpmorig | *.rpmsave | *-range
    devlist = []
    for entry in os.listdir(SYSCONFDEVICEDIR):
        if (len(entry) > 6) and \
           entry.startswith("ifcfg-") and \
           not entry.endswith("~") and \
           not entry.endswith(".bak") and \
           not entry.endswith(".orig") and \
           not entry.endswith(".rpmnew") and \
           not entry.endswith(".rpmorig") and \
           not entry.endswith(".rpmsave") and \
           not entry.endswith("-range") and \
           isfile(SYSCONFDEVICEDIR + entry) and \
           os.access(SYSCONFDEVICEDIR + entry, os.R_OK):
            devlist.append(entry[6:])
    return devlist


def getConfig(device):
    conf = ConfDevice(device)
    return conf.vars


def getBondSlavesFromConfig(device):
    devlist = []
    for dev in getConfiguredDevices():
        conf = ConfDevice(dev)
        if conf.get("MASTER") == device:
            devlist.append(dev)
    return devlist


def getBondMastersFromConfig():
    devlist = []
    for dev in getConfiguredDevices():
        conf = ConfDevice(dev)
        if "MASTER" in conf.keys():
            devlist.append(conf['MASTER'])
    return devlist


def getBridgePortsFromConfig(device):
    devlist = []
    for dev in getConfiguredDevices():
        conf = ConfDevice(dev)
        if conf.get("BRIDGE") == device:
            devlist.append(dev)
    return devlist


def getHostnameFromConfig():
    f = open(HOSTNAMEFILE)
    try:
        return f.read().strip()
    finally:
        f.close()


def ifupDevice(device):
    if os.access(SYSCONFDEVICEDIR + "ifcfg-" + device, os.R_OK):
        shell_cmd("ifup %s boot" % device)


def ifdownDevice(device):
    if device in getDevices() and os.access(SYSCONFDEVICEDIR + "ifcfg-" + device, os.R_OK):
        shell_cmd("ifdown %s" % device)


def activateDevice(device):
    if device in getDevices():
        shell_cmd("ip link set %s up" % device)


def deactivateDevice(device):
    if device in getVlans():
        shell_cmd("vconfig rem %s" % device)
    if isBridgePort(device):
        shell_cmd("brctl delif %s %s" % (getBridge(device), device))
    if device in getBridges():
        for port in getBridgePorts(device):
            shell_cmd("brctl delif %s %s" % (device, port))
        shell_cmd("ip link set %s down" % device)
        shell_cmd("brctl delbr %s" % device)
    if isBondSlave(device):
        shell_cmd("ifenslave -d %s %s" % (getBondMaster(device), device))
    if device in getBonds():
        for slave in getBondSlaves(device):
            shell_cmd("ifenslave -d %s %s" % (device, slave))
        try:
            f = open(BONDING_MASTERS, "w")
            f.write("-" + device)
            f.close()
        except OSError, e:
            raise NetworkError("Failed to deactivate %s: %s" % (device, e))
    if device in getActiveDevices():
        shell_cmd("ip link set %s down" % device)


def createBond(device, slaves, options={}):
    if isinstance(slaves, str):
        slaves = [x.strip() for x in slaves.split(",")]
    # FIXME: Linux bonding driver doesn't require the slaves be physical.
    # But it's safe to add restriction.
    #if not set(slaves).issubset(set(getPhysicalDevices())):
    if not set(slaves).issubset(set(getDevices())):
        raise NetworkError("Slaves should be physical and exist: %s" % slaves)
    if not set(getBondSlaves(device)).issubset(set(slaves)):
        raise NetworkError("Extra slaves exist: %s" % getBondSlaves(device))
    mtu = None
    for slave in slaves:
        if mtu and getMTU(slave) != mtu:
            raise NetworkError("Slaves must have the smae MTU")
        else:
            mtu = getMTU(slave)
    # modprobe does the right thing if the module is already loaded.
    shell_cmd("modprobe bonding")
    if not device in getBonds():
        try:
            f = open(BONDING_MASTERS, "w")
            f.write("+" + device)
            f.close()
        except IOError, e:
            raise NetworkError("Failed to create bond: %s" % e)
    setBondOptions(device, options)
    activateDevice(device)
    for slave in slaves:
        shell_cmd("ip link set %s down" % slave)
        if slave not in getBondSlaves(device):
            shell_cmd("ifenslave %s %s" % (device, slave))
        activateDevice(slave)


def getVlanName(device, vlanid, options={}):
    vlanid = int(vlanid)
    if vlanid < 0 or vlanid >= 4096:
        raise NetworkError("Vlan tag invalvlanid: %s" % vlanid)
    vlan = "%s.%d" % (device, vlanid)
    name_type = options.get('name_type')
    if name_type in ["VLAN_PLUS_VID", "VLAN_PLUS_VID_NO_PAD",
                     "DEV_PLUS_VID", "DEV_PLUS_VID_NO_PAD"]:
        (rc, _, _) = shell_cmd("vconfig set_name_type %s" % name_type)
        if rc == 0:
            if options["name_type"] == "VLAN_PLUS_VID":
                vlan = "vlan%.4d" % vlanid
            elif options["name_type"] == "VLAN_PLUS_VID_NO_PAD":
                vlan = "vlan%d" % vlanid
            elif options["name_type"] == "DEV_PLUS_VID":
                vlan = "%s.%.4d" % (device, vlanid)
            elif options["name_type"] == "DEV_PLUS_VID_NO_PAD":
                vlan = "%s.%d" % (device, vlanid)
    return vlan


def createVlan(device, vlanid, options={}):
    if device not in getDevices():
        raise NetworkError("Device not exist: %s" % device)
    shell_cmd("modprobe 8021q")
    activateDevice(device)
    vlan = getVlanName(device, vlanid, options)
    if vlan not in getVlans():
        (status, stdoutdata, stderrdata) = shell_cmd("vconfig add %s %d" % (device, vlanid))
        if status != 0:
            raise NetworkError("Failed to create vlan: %s" % stderrdata)
    activateDevice(vlan)
    setVlanOptions(vlan, options)


def createBridge(device, options={}):
    if device not in getBridges():
        shell_cmd("brctl addbr %s" % device)
    activateDevice(device)
    setBridgeOptions(device, options)


def setBondOptions(device, options):
    option_path = join(SYSNETDIR, device, "bonding")
    for (option, value) in options.items():
        if exists(join(option_path, option)):
            try:
                f = open(join(option_path, option), "w")
                f.write(str(value))
                f.close()
            except IOError, e:
                raise NetworkError("Failed to set option %s to %s: %s" % (option, value, e))


def setVlanOptions(device, options):
    for (option, value) in options.items():
        if option in ["name_type"] and \
           str(value) in ["VLAN_PLUS_VID", "VLAN_PLUS_VID_NO_PAD",
                          "DEV_PLUS_VID", "DEV_PLUS_VID_NO_PAD"]:
            (rc, _, _) = shell_cmd("vconfig set_%s %s" % (option, value))
        elif option in ["flag", "egress_map", "ingress_map"]:
            (rc, _, _) = shell_cmd("vconfig set_%s %s %s" % (option, device, value))
        if rc != 0:
            raise NetworkError("Failed to set option %s to %s" % (option, value))


def setBridgeOptions(device, options):
    for (option, value) in options.items():
        if option in ["stp"]:
            (rc, _, _) = shell_cmd("brctl stp %s %s" % (device, value))
        elif option in ["ageing", "bridgeprio", "fd", "hello", "maxage",
                        "pathcost", "ortprio"]:
            (rc, _, _) = shell_cmd("brctl set%s %s %s" % (option, device, value))
        if rc != 0:
            raise NetworkError("Failed to set option: %s to %s" % (option, value))


def setEtherToolOptions(device, options):
    # Options for "ethtool -s": ["autoneg", "speed", "duplex"]
    settings = ""
    # Options for "ethtool -K"
    offloads = ""
    offload_opts = ["rx", "tx", "sg", "tso", "ufo", "gso"]

    for (option, value) in options.items():
        if option in ["speed"]:
            if value in ["10", "100", "1000"]:
                settings = " ".join([settings, option, value])
            else:
                raise NetworkError("Invalid value for ethtool speed: %s. "
                         "Must be 10|100|1000." % value)
        elif option in ["duplex"]:
            if value in ["half", "full"]:
                settings = " ".join([settings, option, value])
            else:
                raise NetworkError("Invalid value for ethtool duplex: %s. "
                         "Must be half|full." % value)
        elif option in ["autoneg"]:
            if value in ["true", "on"]:
                settings = " ".join([settings, option, "on"])
            elif value in ["false", "off"]:
                settings = " ".join([settings, option, "off"])
            else:
                raise NetworkError("Invalid value for ethtool autoneg: %s. "
                         "Must be on|true|off|false." % value)
        elif option in offload_opts:
            if value in ["true", "on"]:
                offloads = " ".join([offloads, option, "on"])
            elif value in ["false", "off"]:
                offloads = " ".join([offloads, option, "off"])
            else:
                raise NetworkError("Invalid value for ethtool %s: %s. "
                         "Must be on|true|off|false." % (option, value))

    if settings:
        (rc, _, _) = shell_cmd("ethtool -s %s %s" % settings)
        if rc != 0:
            raise NetworkError("Failed to change settings: %s %s" % (device, settings))

    if offloads:
        (rc, _, _) = shell_cmd("ethtool -K %s %s" % (device, offloads))
        if rc != 0:
            raise NetworkError("Failed to change offloads: %s" % offloads)


def getDevices():
    devlist = []
    if exists(SYSNETDIR):
        for dev in os.listdir(SYSNETDIR):
            if isdir(join(SYSNETDIR, dev)):
                devlist.append(dev)
    return devlist


def getHardware(device):
    if device in getPhysicalDevices():
        try:
            dev_path = os.readlink(join(SYSNETDIR, device, "device"))
            if not re.search("pci", dev_path):
                return
            return basename(dev_path)
        except OSError, e:
            raise NetworkError("Read link failed: %s" % e)


def getActiveDevices():
    devlist = []
    for dev in getDevices():
        operstate = join(SYSNETDIR, dev, "operstate")
        if not exists(operstate):
            # No operstate for interface assuming down
            continue
        f = open(operstate)
        state = f.read().strip()
        f.close()
        # bridges confusingly return "unknown" when they are up
        if state in ["up", "unknown"]:
            devlist.append(dev)
    return devlist


def getPhysicalDevices():
    devlist = []
    for dev in getDevices():
        if exists(join(SYSNETDIR, dev, "device")):
            devlist.append(dev)
    return devlist


def getBonds():
    devlist = []
    if os.access(BONDING_MASTERS, os.F_OK | os.R_OK) and isfile(BONDING_MASTERS):
        f = open(BONDING_MASTERS, "r")
        bonds = f.readline().strip()
        f.close()
        devlist = bonds.split()
    return devlist


def getBondSlaves(device):
    devlist = []
    for dev in getDevices():
        if exists(join(SYSNETDIR, device, "slave_%s" % dev)):
            devlist.append(dev)
    return devlist


def getBondMaster(device):
    if isBondSlave(device):
        return basename(os.readlink(join(SYSNETDIR, device, "master")))


def isBondSlave(device):
    return exists(join(SYSNETDIR, device, "master"))


def isBridgePort(device):
    return exists(join(SYSNETDIR, device, "brport"))


def getBridge(device):
    if isBridgePort(device):
        return basename(os.readlink(join(SYSNETDIR, device, "brport", "bridge")))


def getBridges():
    devlist = []
    (status, stdoutdata, stderrdata) = shell_cmd('/usr/sbin/brctl show | cut -d "\n" -f 2- | cut -f 1')
    if status == 0:
        devlist = stdoutdata.split()
    return devlist


def getBridgePorts(device):
    devlist = []
    brif_dir = join(SYSNETDIR, device, "brif")
    if exists(brif_dir) and isdir(brif_dir):
        devlist = os.listdir(brif_dir)
    return devlist


def getVlans():
    devlist = []
    for device in getDevices():
        if exists("/proc/net/vlan/%s" % device):
            devlist.append(device)
    return devlist


def _setParamSysfs(device, key, value):
    try:
        f = open(join(SYSNETDIR, device, key), "w")
        f.write(value)
        f.close()
    except OSError, e:
        raise NetworkError("Failed to set %s to %s for %s: %s" % (key, value, device, e))


def _setParamIP(device, key, value):
    cmd = "ip link set %s %s %s" % (device, key, value)
    (status, stdoutdata, stderrdata) = shell_cmd(cmd)
    if status != 0:
        raise NetworkError("Failed to set %s to %s for %s: %s" % (key, value, device, stderrdata))


def _getParamSysfs(device, key):
    f = open(join(SYSNETDIR, device, key))
    value = f.read()
    f.close()
    return value.strip()


def setMTU(device, value):
    _setParamSysfs(device, "mtu", value)


def setMAC(device, value):
    _setParamIP(device, "addr", value)


def calcDottedNetmask(mask):
    bits = 0xffffffff ^ (1 << 32 - int(mask)) - 1
    return socket.inet_ntoa(struct.pack('>I', bits))


def getAddr(device):
    cmd = "ip -o -f inet addr show dev %s label %s" % (device, device)
    (status, stdoutdata, stderrdata) = shell_cmd(cmd)
    if status == 0:
        for line in stdoutdata.split("\n"):
            m = re.match(r"\d+:\s+%s\s+inet\s+(?P<ipaddr>\d+\.\d+\.\d+\.\d+)/(?P<netmask>\d+)\s+brd\s.*" % device, line)
            if m:
                return m.group("ipaddr"), calcDottedNetmask(m.group('netmask'))


def getMTU(device):
    return _getParamSysfs(device, "mtu")


def getMAC(device):
    return _getParamSysfs(device, "address")


def getStatistics(device):
    stat_dir = join(SYSNETDIR, device, "statistics")
    stat = {}
    for key in os.listdir(stat_dir):
        f = open(join(stat_dir, key))
        value = f.read()
        f.close()
        stat[key] = value.strip()
    return stat


def getHostname():
    return socket.gethostname()


def setHostname(name):
    status, stdoutdata, stderrdata = shell_cmd('hostnamectl set-hostname "%s"' % name)
    if status != 0:
        raise NetworkError('Failed to set hostname to %s: (%s) %s' % (name, status, stderrdata))


def setHosts(ipaddr, hostname, *aliases):
    f = open(HOSTSFILE, "r+")
    content = f.read()
    entry = '    '.join((ipaddr, hostname) + aliases)
    if re.search(r'(^|\n)%s\s+' % ipaddr, content):
        content = re.sub(r'(^|\n)%s\s+.*' % ipaddr, r'\g<1>%s' % entry, content)
    else:
        content += "\n%s\n" % entry
    f.seek(0)
    f.truncate(0)
    f.write(content)
    f.close()
